// ignore_for_file: avoid_print, deprecated_member_use_from_same_package

import 'package:flutter/material.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:wiredash/src/_wiredash_internal.dart';
import 'package:wiredash/src/feedback/data/feedback_item.dart';
import 'package:wiredash/wiredash.dart';

import 'util/robot.dart';
import 'util/wiredash_tester.dart';

void main() {
  test('userEmail can be set', () async {
    final WiredashModel model = WiredashModel(createMockServices());
    final controller = WiredashController(model);
    expect(controller.metaData.userEmail, isNull);
    controller.setUserProperties(userEmail: 'dash@flutter.io');
    expect(controller.metaData.userEmail, 'dash@flutter.io');
  });

  test('userEmail can be resetted', () async {
    final WiredashModel model = WiredashModel(createMockServices());
    final controller = WiredashController(model);
    expect(controller.metaData.userEmail, isNull);
    controller.setUserProperties(userEmail: 'dash@flutter.io');
    expect(controller.metaData.userEmail, 'dash@flutter.io');
    controller.setUserProperties(userEmail: null);
    expect(controller.metaData.userEmail, isNull);
  });

  test('custom metadata can not be mutable via metaData getter', () async {
    final WiredashModel model = WiredashModel(createMockServices());
    final controller = WiredashController(model);
    final map = controller.metaData.custom;
    expect(() => map['foo'] = 'bar', throwsA(isA<UnsupportedError>()));
  });

  test('custom metadata can be mutated in modifyMetaData', () async {
    final WiredashModel model = WiredashModel(createMockServices());
    final controller = WiredashController(model);
    final map = controller.metaData.custom;
    controller.modifyMetaData((metaData) => metaData..custom['foo'] = 'bar');
    expect(controller.metaData.custom['foo'], 'bar');
    // getter copy did not change
    expect(map['foo'], isNull);
  });

  testWidgets('build information can be set via ENV', (tester) async {
    final robot = WiredashTestRobot(tester);
    await robot.launchApp();
    robot.services.inject<BuildInfo>((_) {
      return const BuildInfo(
        compilationMode: CompilationMode.profile,
        // fake from ENV
        buildCommit: 'abc',
        buildNumber: '123',
        buildVersion: '1.2.3',
      );
    });
    // send ping
    await tester.pumpSmart(const Duration(seconds: 5));
    final latestPing = robot.mockServices.mockApi.pingInvocations.latest;
    final ping = latestPing[0] as PingRequestBody?;
    expect(ping!.buildVersion, '1.2.3');
    expect(ping.buildNumber, '123');
    expect(ping.buildCommit, 'abc');
  });

  testWidgets('empty strings from build information ENVs are ignored',
      (tester) async {
    final robot = WiredashTestRobot(tester);
    await robot.launchApp();
    robot.services.inject<BuildInfo>((_) {
      return const BuildInfo(
        compilationMode: CompilationMode.profile,
        // fake from ENV
        buildCommit: '',
        buildNumber: '',
        buildVersion: '',
      );
    });
    // send ping
    await tester.pumpSmart(const Duration(seconds: 5));
    final latestPing = robot.mockServices.mockApi.pingInvocations.latest;
    final ping = latestPing[0] as PingRequestBody?;
    expect(ping!.buildVersion, '0.1.0'); // from appInfo
    expect(ping.buildNumber, '1'); // from appInfo
    expect(ping.buildCommit, isNull);
  });

  testWidgets('set metadata before opening wiredash', (tester) async {
    final robot = WiredashTestRobot(tester);

    CustomizableWiredashMetaData? metadata;
    await robot.launchApp(
      builder: (context) {
        return Scaffold(
          body: Column(
            children: [
              GestureDetector(
                onTap: () async {
                  final wiredash = Wiredash.of(context);
                  wiredash.modifyMetaData((metaData) {
                    metadata = metaData.copyWith();
                    return metaData;
                  });
                  wiredash.show();
                },
                child: const Text('Feedback'),
              ),
            ],
          ),
        );
      },
    );
    await robot.openWiredash();

    // can only be set by developers
    expect(metadata!.userId, isNull);
    expect(metadata!.userEmail, isNull);
    expect(metadata!.custom.isEmpty, isTrue);
  });

  testWidgets('reading metadata as first wiredash interaction', (tester) async {
    final robot = WiredashTestRobot(tester);

    CustomizableWiredashMetaData? metadata;
    await robot.launchApp(
      builder: (context) {
        return Scaffold(
          body: Column(
            children: [
              GestureDetector(
                onTap: () async {
                  final wiredash = Wiredash.of(context);
                  metadata = wiredash.metaData;
                  wiredash.show();
                },
                child: const Text('Feedback'),
              ),
            ],
          ),
        );
      },
    );
    await robot.openWiredash();

    // all empty
    expect(metadata!.userId, isNull);
    expect(metadata!.userEmail, isNull);
    expect(metadata!.custom.isEmpty, isTrue);
  });

  group('collectMetaData', () {
    testWidgets(
        'Wiredash.collectMetaData has precedence over WiredashFeedbackOptions',
        (tester) async {
      final robot = WiredashTestRobot(tester);
      await robot.launchApp(
        collectMetaData: (metaData) {
          return metaData..userEmail = "primary@mail.com";
        },
        feedbackOptions: WiredashFeedbackOptions(
          collectMetaData: (metaData) {
            return metaData..userEmail = "secondary@mail.com";
          },
        ),
      );
      await robot.submitMinimalFeedback();
      final latestCall =
          robot.mockServices.mockApi.sendFeedbackInvocations.latest;
      final submittedFeedback = latestCall[0] as FeedbackItem?;
      expect(submittedFeedback!.metadata.userEmail, 'primary@mail.com');
    });

    testWidgets('Use WiredashFeedbackOptions.collectMetaData as fallback',
        (tester) async {
      final robot = WiredashTestRobot(tester);
      await robot.launchApp(
        feedbackOptions: WiredashFeedbackOptions(
          collectMetaData: (metaData) {
            return metaData..userEmail = "secondary@mail.com";
          },
        ),
      );
      await robot.submitMinimalFeedback();
      final latestCall =
          robot.mockServices.mockApi.sendFeedbackInvocations.latest;
      final submittedFeedback = latestCall[0] as FeedbackItem?;
      expect(submittedFeedback!.metadata.userEmail, 'secondary@mail.com');
    });

    testWidgets('Wiredash.collectMetaData has precedence over PsOptions',
        (tester) async {
      final robot = WiredashTestRobot(tester);
      await robot.launchApp(
        collectMetaData: (metaData) {
          return metaData..userEmail = "primary@mail.com";
        },
        psOptions: PsOptions(
          collectMetaData: (metaData) {
            return metaData..userEmail = "secondary@mail.com";
          },
        ),
      );
      await robot.openPromoterScore();
      await tester.pump(const Duration(seconds: 1));
      final latestCall = robot.mockServices.mockApi.sendPsInvocations.latest;
      final ps = latestCall[0] as PromoterScoreRequestBody?;
      expect(ps!.metadata.userEmail, 'primary@mail.com');
    });

    testWidgets('Use PsOptions.collectMetaData as fallback', (tester) async {
      final robot = WiredashTestRobot(tester);
      await robot.launchApp(
        psOptions: PsOptions(
          collectMetaData: (metaData) {
            return metaData..userEmail = "secondary@mail.com";
          },
        ),
      );
      await robot.openPromoterScore();
      await tester.pump(const Duration(seconds: 1));
      final latestCall = robot.mockServices.mockApi.sendPsInvocations.latest;
      final ps = latestCall[0] as PromoterScoreRequestBody?;
      expect(ps!.metadata.userEmail, 'secondary@mail.com');
    });

    testWidgets('All user collected metaData is submitted to the server',
        (tester) async {
      final robot = WiredashTestRobot(tester);
      await robot.launchApp(
        collectMetaData: (metaData) {
          return metaData
            ..userEmail = "user@mail.com"
            ..userId = "123"
            ..custom['foo'] = 'bar';
        },
      );
      await robot.submitMinimalFeedback();
      final latestCall =
          robot.mockServices.mockApi.sendFeedbackInvocations.latest;
      final submittedFeedback = latestCall[0] as FeedbackItem?;
      expect(submittedFeedback!.metadata.userEmail, 'user@mail.com');
      expect(submittedFeedback.metadata.userId, '123');
      expect(submittedFeedback.metadata.custom!['foo'], 'bar');
    });
  });
  group('sessionMetaData', () {
    testWidgets('sent along feedback', (tester) async {
      final robot = WiredashTestRobot(tester);
      await robot.launchApp();
      await robot.submitMinimalFeedback();
      final latestCall =
          robot.mockServices.mockApi.sendFeedbackInvocations.latest;
      final submittedFeedback = latestCall[0] as FeedbackItem?;
      expect(submittedFeedback!.metadata.appBrightness, Brightness.light);
      expect(submittedFeedback.metadata.appLocale, 'en-US');
    });

    testWidgets('send along promoter score', (tester) async {
      final robot = WiredashTestRobot(tester);
      await robot.launchApp();
      await robot.openPromoterScore();
      await tester.pump(const Duration(seconds: 1));
      final latestCall = robot.mockServices.mockApi.sendPsInvocations.latest;
      final ps = latestCall[0] as PromoterScoreRequestBody?;
      expect(ps!.metadata.appBrightness, Brightness.light);
      expect(ps.metadata.appLocale, 'en-US');
    });
  });
}
